#!/usr/bin/lua
local socket
local encapGRE
local decapGRE
local createGRE
local createTun
local tunnels
local int32ToStr
local strToInt32
local int16ToStr
local strToInt16
local int8ToStr
local strToInt8
local sendKeepAlive
local getPath
local setPath
local nsg
local defaultTtl
local tunSend
local splitSend
local checkBuf
local insertBuf

socket=require"socket"
nsg = require"nsg"
require"bit"
local bit = bit

tunnels={
   max = 0,
   hash = {},
   KAs = {},
   KATime = 0,
   names = {},
   Bs = {},
   BTime = 0,
   stats = {
      enfprx = 0, -- tunnel Not Found
      ebpprx = 0, -- can`t decapsulate  , crc or short packet
   }
}

local greConfig = {

}
controlCommands = {
   
}

sendKeepAlive = function(tun)
                   local tunCfg = tunnels[tun]
                   local gre = tunCfg.link
                   --local greCfg = tunnels[gre]
                   --io.write("KAE:"..tunCfg.KAError.."KAW:"..tostring(tunCfg.KAWait).."\n")
                   if tunCfg.KAWait then
                      tunCfg.KAError = tunCfg.KAError + 1
                      if (tunCfg.KAError > tunCfg.KARetry)
                      and (tunCfg.KAUP or math.fmod(tunCfg.KAError, 10)
 )
                      then
                         --io.write("keepalive!!! DOWN IFACE:"..tunCfg.name.."\n")
                         tunCfg.KAUP = false
                         io.write(tunCfg.name.." down\n")
                         os.execute("ifconfig "..tunCfg.name.." down")
                      end
                   end
                   local ret, err = gre:sendto(tunCfg.KAPacket, tunCfg.destination, 0)
                   local stats = tunCfg.stats
                   if not ret then
                      io.write("err".."\n")
                      stats.elkaptx = stats.elkaptx + 1
                   else
                      stats.lkatx = stats.lkatx + 20 + #(tunCfg.KAPacket)
                      stats.lkaptx = stats.lkaptx + 1
                   end
                   tunCfg.KAWait = true
                end

int8ToStr = function(int)
               int = tonumber(int)
               return string.char(int % 256)
            end

int16ToStr = function(int)
                int = tonumber(int)
                return string.char(math.floor(int/2^8) % 256, int % 256)
             end

int32ToStr = function(int)
                int = tonumber(int)
                return string.char(
                   math.floor(int/2^24) % 256,
                   math.floor(int/2^16) % 256,
                   math.floor(int/2^8) % 256,
                   int % 256)
             end

strToInt8 = function(str)
                b1 = str:byte(1)
                return (b1)
             end

strToInt16 = function(str)
                b1,b2 = str:byte(1,2)
                return (b2 + b1 * 2^8)
             end
strToInt32 = function(str)
                b1,b2,b3,b4 = str:byte(1,4)
                return (b4 + b3 * 2^8 + b2 * 2^16 + b1 * 2^24)
             end


encapGRE = function (buf, tun, sequence)
              local res
              if sequence then
                 res = table.concat({
                                       tun.GREHeader.flagsVer,
                                       tun.GREHeader.proto,
                                       tun.GREHeader.checksum,
                                       tun.GREHeader.key,
                                       sequence and int32ToStr(sequence) or "",
                                       buf,
                                    })
              else
                 res = table.concat({tun.GREHeader.head, buf})
              end
              --io.write("head crc:"..table.concat({tun.GREHeader.head:byte(1,#tun.GREHeader.head)}," ")..":"..(nsg.ipsum(tun.GREHeader.head, #tun.GREHeader.head)).."\n")

              if tun.checksum then
                 local crc = nsg.ipsum(res, #res)
                 res =  table.concat({
                                        tun.GREHeader.flagsVer,
                                        tun.GREHeader.proto,
                                        int16ToStr(crc),
                                        int16ToStr(0), --reserved
                                        tun.GREHeader.key,
                                        sequence and int32ToStr(sequence) or "",
                                        buf,
                                     })
                 --io.write("crc:"..(nsg.ipsum(res, #res)).."\n")
              end
              return res
           end

decapGRE = function (buf, size)
              if not buf or #buf < 24 then
                 -- io.write("Error receiving data, data is too short.")
                 return nil
              end
              local IPVersion = math.floor(buf:byte(1)/16)
              local IPHeaderLen
              local key = ""
              local checksum
              local encap = 0
              local seq

              if IPVersion == 4 then
                 IPHeaderLen = 4 * (buf:byte(1) - IPVersion * 16)
              elseif IPVersion ==  6 then
                 IPHeaderLen = 40
              end
              local GREFlags = buf:byte(1 + IPHeaderLen)
              encap = strToInt16(buf:sub(1 + IPHeaderLen + 2, 1 + IPHeaderLen + 3))
              local GREHeaderLen = 4
              if  (bit.band(GREFlags, 128) == 128) and (#buf >= 1 + IPHeaderLen + 3) then  -- checksum
                 local grePacket = buf:sub(1 + IPHeaderLen)
                 if 0 ~=  nsg.ipsum(grePacket, #grePacket) then
                    --io.write("Bad GRE checksum, drop packet")
                    return nil
                 end
                 checksum = strToInt16(buf:sub(1 + IPHeaderLen + GREHeaderLen ,2 + IPHeaderLen + GREHeaderLen))
                 GREHeaderLen = GREHeaderLen + 4
              end
              if (bit.band(GREFlags, 32) == 32) and (#buf >= 4 + IPHeaderLen + GREHeaderLen) then --key
                 key = strToInt32(buf:sub(1 + IPHeaderLen + GREHeaderLen, 4 + IPHeaderLen + GREHeaderLen ))
                 GREHeaderLen = GREHeaderLen + 4
              end
              if (bit.band(GREFlags, 16) == 16) and (#buf >= 4 + IPHeaderLen + GREHeaderLen) then --seq
                 seq = strToInt32(buf:sub(1 + IPHeaderLen + GREHeaderLen, 4 + IPHeaderLen + GREHeaderLen ))
                 GREHeaderLen = GREHeaderLen + 4
              end
              if #buf < IPHeaderLen + GREHeaderLen then
                 return nil
              end
              buf=buf:sub(1 + IPHeaderLen + GREHeaderLen) or ""
              return buf, encap, checksum , key, seq
           end

createCTL = function (addr, port)
               local sockCTL, err = socket.udp()
               if err then
                  io.write(err)
               else
                  sockCTL:settimeout(0)
                  sockCTL:setsockname(addr or "127.0.0.1", port or 50009)
                  tunnels[sockCTL] = {
                     socket = sockCTL,
                     type = "CTL",
                     name = "CTL",
                     index = tunnels.max + 1,
                  }
                  tunnels.max = tunnels.max + 1
                  tunnels[#tunnels+1] = sockCTL
                  end
                  return sockCTL, err
               end

getPath = function (root, first, ...)
                return type(root)=="table" and first and getPath(root[first], ...) or 
             type(root)~="table" and root
             end

setPath = function (root, value, ...)
             local prevRoot
             local index
             for _, v in ipairs({...}) do
                prevRoot = root
                index = v
                root[v] = root[v] or {}
                root = root[v]
             end
             prevRoot[index] = value
          end

deleteGRE = function (source, device)
               device = device or ""
               source = source or "0.0.0.0"
               local sockGRE, err
               sockGRE = getPath(tunnels.hash, device, source, "GRE")
               if sockGRE then
                  tunnels[sockGRE].usage =  tunnels[sockGRE].usage - 1
                  if tunnels[sockGRE].usage == 0 then
                     setPath(tunnels.hash, nil, device, source, "GRE")
                     local GRECfg = tunnels[sockGRE]
                     tunnels[GRECfg.index] =  tunnels[#tunnels]
                     tunnels[tunnels[GRECfg.index]].index = GRECfg.index
                     tunnels[sockGRE] = nil
                     tunnels[#tunnels] = nil
                     sockGRE:close()
                     tunnels.max = tunnels.max - 1
                  end
               end
            end

createGRE = function (source, device)
               device = device or ""
               source = source or "0.0.0.0"
               local sockGRE, err
               sockGRE = getPath(tunnels.hash, device, source, "GRE")
               if sockGRE then
                  tunnels[sockGRE].usage =  tunnels[sockGRE].usage + 1
                  return sockGRE
               end
               sockGRE, err = socket.raw(47) --47 for gre
               if err then
                  return nil, err
               else
                  setPath(tunnels.hash, sockGRE, device, source, "GRE")
                  sockGRE:settimeout(0)
                  sockGRE:setsockname(source)
                  tunnels[sockGRE] = {
                     socket = sockGRE,
                     type = "GRE",
                     name = "GRE:"..source,
                     source = source,
                     device = device,
                     index = tunnels.max + 1,
                     usage = 1,
                  }
                  tunnels.max = tunnels.max + 1
                  tunnels[#tunnels+1] = sockGRE
               end
               return sockGRE
            end

deleteTun = function (device, source, destination, key)
               local sockTun = getPath(tunnels.hash, device, source, destination, key or "")
               setPath(tunnels.hash, nil, device, source, destination, key or "")
               if sockTun then
                  local TunCfg = tunnels[sockTun]
                  tunnels.KAs[sockTun] = nil
                  tunnels[TunCfg.index] =  tunnels[#tunnels]
                  tunnels[tunnels[TunCfg.index]].index = TunCfg.index
                  tunnels[sockTun] = nil
                  tunnels[#tunnels] = nil
                  sockTun:close()
                  tunnels.max = tunnels.max - 1
               end
            end

local maxFd = 1024
createTun = function (name, link, source, destination, encap, key , checksum, KAPeriod, KARetry, sequence, sequenceBuffer, sequenceTimer, ttl, tos, syncMax, syncTimer, syncIntegrity)

               if tunnels.max+50 > 1024 then
                  if socket.select ~= socket.poll then
                     socket.select = socket.poll
                     io.write("Use poll instead of select\n")
                  end
                  maxFd = (maxFd < tunnels.max+50) and (tunnels.max+50) or maxFd
                  socket.setrlimit(nil, maxFd, maxFd)
               end

               --io.write( ser.serialize(tunnels.hash,2))
               if  getPath(tunnels.hash, tunnels[link].device, tunnels[link].source, destination, key or "") then
                  return nil, "Tunnel '"..tunnels[getPath(tunnels.hash, tunnels[link].device, tunnels[link].source, destination, key or "")].name.."' has the same config."
               end
               local  tun, err = socket.tun(name, encap=="eth" and 1 or nil)
               if err then
                  return nil, err
               else
                  tun:settimeout(0)
                  tunnels.names[name] = tun
                  tunnels[tun] = {
                     socket = tun,
                     type = "tun",
                     link = link,
                     name = name,
                     encap = encap=="eth" and 25944 or 2048,
                     destination = destination,
                     key = key,
                     ttl = tonumber((tostring(ttl)):match("[%d]+") or defaultTtl) or 64,
                     tos = tos or 0,
                     sequence = sequence and {
                        txSeq = 0,
                        rxSeq = (2 ^ 32) -1,
                        bufMax = sequenceBuffer,
                        bufTimer = sequenceTimer/1000,
                        bufs = {next = nil},
                        bufLen = 0,
                        syncs = {next = nil},
                        syncMax = (syncMax or 0),
                        syncTimer = (syncTimer or 0)/1000,
                        syncIntegrity = (syncIntegrity or 0),
                        syncLen = 0,
                     } or nil,
                     checksum = (checksum=="yes") and true or false,
                     KAPeriod = KAPeriod,
                     KARetry = KARetry,
                     index = tunnels.max + 1,
                     stats={rx=0, tx=0, prx = 0 , ptx = 0, eprx = 0 , eptx = 0,
                            lkatx = 0, lkarx = 0, rkatx = 0, rkarx = 0,
                            lkaptx = 0, lkaprx = 0, rkaptx = 0, rkaprx = 0,
                            elkaptx = 0, elkaprx = 0, erkaptx = 0, erkaprx = 0},
                  }
                  local tunCfg = tunnels[tun]
                  local greCfg = tunnels[link]
                  tunCfg.GREHeader = {
                     flagsVer = string.char((tunCfg.checksum and 128 or 0) + (key and 32 or 0) +(sequence and 16 or 0), 0),
                     proto = int16ToStr(encap=="eth" and 25944 or 2048) ,
                     checksum = tunCfg.checksum and int32ToStr(0) or "",
                     key = key and int32ToStr(tonumber(key)) or "",
                     sequence = sequence and  int32ToStr(0) or ""
                  }
                  
                  tunCfg.GREHeader.head = table.concat({
                            tunCfg.GREHeader.flagsVer,
                            tunCfg.GREHeader.proto,
                            tunCfg.GREHeader.checksum,
                            tunCfg.GREHeader.key,
                            tunCfg.GREHeader.sequence,

                                                       })
                  tunCfg.GREHeader.headLen = #tunCfg.GREHeader.head

                  setPath(tunnels.hash, tun, greCfg.device, greCfg.source, destination, key or "" )
                  
                  tunnels.max = tunnels.max + 1
                  tunnels[#tunnels+1] =  tun
                  if (KAPeriod and (type(KAPeriod) == "number") and (KAPeriod >0) and 
                KARetry  and (type(KARetry) == "number") and (KARetry >0)) then
                     tunnels.KAs[tun] = tunCfg
                     tunCfg.KAUP = false
                     tunCfg.KAWait = true
                     tunCfg.KAError = 9999999
                     tunCfg.KALastTime = 0
                     tunCfg.KASource = strToInt32(string.char(socket.skip(2,(socket.dns.toip(destination) or destination ):find("(%d+)%.(%d+)%.(%d+)%.(%d+)"))))
                     tunCfg.KADestination = strToInt32(string.char(socket.skip(2,(socket.dns.toip(source) or source):find("(%d+)%.(%d+)%.(%d+)%.(%d+)"))))

--using of KEY in KA packet is non standard!!!
                     local KAGRE  =  string.char((key and 32 or 0),0,0,0)..(key and int32ToStr(key) or "")
                     local KAIPHead =  table.concat(
                        {
                           "\69",--size ver + headLen
                           int8ToStr(tonumber(tunCfg.tos) or 0), --tos
                           int16ToStr(20+#KAGRE), --total size
                           "\0\0", --ident
                           "\0\0", --flags + frag
                           int8ToStr(tunCfg.ttl), --ttl
                           "\47", --proto
                           "\0\0",  --check sum
                           int32ToStr(tunCfg.KASource),--source
                           int32ToStr(tunCfg.KADestination),--destination
                        }
                     )
                     tunCfg.KAPacket = table.concat(
                        {
                           string.char((key and 32 or 0),0), -- gre flags + ver
                           int16ToStr(2048), -- encap proto IP
                           (key and int32ToStr(key) or ""), -- GRE key
                           "\69",--size ver + headLen
                           int8ToStr(tonumber(tunCfg.tos) or 0), --tos
                           int16ToStr(20+#KAGRE), --total size
                           "\0\0", --ident
                           "\0\0", --flags + frag
                           int8ToStr(tunCfg.ttl), --ttl
                           "\47", --proto
                           int16ToStr(nsg.ipsum(KAIPHead, #KAIPHead)),  --check sum
                           int32ToStr(tunCfg.KASource),--source
                           int32ToStr(tunCfg.KADestination),--destination
                           KAGRE,
                        }
                     )
                     os.execute("ifconfig "..tunCfg.name.." down")
                  else
                     tunCfg.KAUP = true
                     io.write  (tunCfg.name.." up\n")
                     os.execute("ifconfig "..tunCfg.name.." up")
                  end
               end
               return tun, err
            end

controlCommands.statistic =
   function ( name )
      if name and name ~= "" then
         if tunnels.names[name] and tunnels[tunnels.names[name]] then
            local stats = tunnels[tunnels.names[name]].stats or {}
            local out = {}
            out[#out + 1] = "RX:\ttotal\tdata\tselfKA\tremoteKA"
            out[#out + 1] = "\t"..(stats.rx+stats.lkarx+stats.rkarx)..":"..(stats.prx+stats.lkaprx+stats.rkaprx).."\t"..stats.rx..":"..stats.prx.."\t"..stats.lkarx..":"..stats.lkaprx.."\t"..stats.rkarx..":"..stats.rkaprx
            out[#out + 1] = "RX Err:\ttotal\tdata\tselfKA\tremoteKA"
            out[#out + 1] = "\t"..(stats.eprx+stats.elkaprx+stats.erkaprx).."\t"..stats.eprx.."\t"..stats.elkaprx.."\t"..stats.erkaprx
            
            out[#out + 1] = "TX:\ttotal\tdata\tselfKA\tremoteKA"
            out[#out + 1] = "\t"..(stats.tx+stats.lkatx+stats.rkatx)..":"..(stats.ptx+stats.lkaptx+stats.rkaptx).."\t"..stats.tx..":"..stats.ptx.."\t"..stats.lkatx..":"..stats.lkaptx.."\t"..stats.rkatx..":"..stats.rkaptx
            out[#out + 1] = "TX Err:\ttotal\tdata\tselfKA\tremoteKA"
            out[#out + 1] = "\t"..(stats.eptx+stats.elkaptx+stats.erkaptx).."\t"..stats.eptx.."\t"..stats.elkaptx.."\t"..stats.erkaptx
            return table.concat(out, "\n").."\n"
         else
            return "Error: no such tunnel:"..name
         end
      else
         local out = {}
         local head = {}
         local tRx,tTx,tpRx,tpTx,tepRx,tepTx = 0,0,0,0,0,0
         out[#out + 1] = "\nTunnel stats"
         for name,_ in pairs(tunnels.names or {}) do
            local stats = tunnels[tunnels.names[name]].stats or {}
            out[#out + 1] = name
            out[#out + 1] ="\tRx:"..(stats.rx+stats.lkarx+stats.rkarx)..":"..(stats.prx+stats.lkaprx+stats.rkaprx).."\tTx:"..(stats.tx+stats.lkatx+stats.rkatx)..":"..(stats.ptx+stats.lkaptx+stats.rkaptx)
            out[#out + 1] ="\tErRx:"..(stats.eprx+stats.elkaprx+stats.erkaprx).."\tErTx:"..(stats.eptx+stats.elkaptx+stats.erkaptx)
            
            tRx = tRx + (stats.rx+stats.lkarx+stats.rkarx)
            tpRx = tpRx + (stats.prx+stats.lkaprx+stats.rkaprx)

            tTx = tTx + (stats.tx+stats.lkatx+stats.rkatx)
            tpTx = tpTx + (stats.ptx+stats.lkaptx+stats.rkaptx)

            tepRx = tepRx + (stats.eprx+stats.elkaprx+stats.erkaprx)
            tepTx = tepTx + (stats.eptx+stats.elkaptx+stats.erkaptx)
           
         end
         head[#head + 1] = "Summary tunnel stats"
         head[#head + 1] = "\tRx:"..tRx..":"..tpRx.."\tTx:"..tTx..":"..tpTx
         head[#head + 1] = "\tErRx:"..tepRx.."\tErTx:"..tepTx
         head[#head + 1] = "\nProtocol errors"
         head[#head + 1] = "\tRx Err No Such Tun:"..tunnels.stats.enfprx
         head[#head + 1] = "\tRx Err Bad Packet(CRC,Len):"..tunnels.stats.ebpprx
         head[#head + 1] = ""
         return  table.concat(head, "\n")..table.concat(out, "\n").."\n"
      end
   end

controlCommands.config = function()
                            return ser.serialize(greConfig,2).."\n"
                      end
controlCommands.addTunnel = function(name, cfg)
                               if not name or not cfg then
                                  return "Error: wrong argument list for addTunnel"
                               else
                                  local oldConf = greConfig[name]
                                  if oldConf
                                  then
                                     return "Warn: tunnel "..name.." already exists"
                                  else
                                     --
                                  end
                                  if cfg.device and #cfg.device>0 and 0~=os.execute("ifconfig "..cfg.device.." 2>/dev/null | grep -o 'UP' >/dev/null") then
                                     return "Warn: no active master device for "..name
                                  end
                                  local gre, tun, err
                                  gre, err = createGRE(cfg.source, cfg.device, cfg.ttl, cfg.tos)
                                  if gre then
                                     tun, err = createTun(name, gre, cfg.source or "0.0.0.0", cfg.destination or "0.0.0.0", cfg.encapsulation, cfg.key, cfg.checksum, cfg.keepalive=="yes" and cfg["keepalive-interval"] or nil,cfg["keepalive-retry"], cfg["sequence-datagrams"]=="yes" and true or nil, cfg["reoder-buffer"], cfg["reoder-timeout"], cfg["ttl"], cfg["tos"], cfg["sync-buffer"], cfg["sync-timeout"],cfg["sync-integrity"] )
                                     if err then
                                        deleteGRE(cfg.source, cfg.device)
                                        return "Error:"..err
                                     else
                                        tunnels[gre].tunCount = (tunnels[gre].tunCount or 0) + 1
                                        greConfig[name] = cfg
                                        --return "OK"
                                     end
                                  else
                                     return "Error:"..err
                                  end
                               end
                               return "OK"
                            end

controlCommands.delTunnel = function(name)
                               local cfg = greConfig[name]
                               if not name or not cfg then
                                  return "Warn: no such gre tunnel"
                               else
                                  greConfig[name] = nil
                                  deleteTun(cfg.device or "",
                                            cfg.source or "0.0.0.0",
                                            cfg.destination or "0.0.0.0",
                                            cfg.key or "")
                                  deleteGRE(cfg.source, cfg.device)
                               end
                               return "OK"
                            end

controlCommands.free = function()
                          os.exit()
                       end

splitSend = function (sock, msg, addr, port)
   local bufSize = 1024
   local l = #msg
   local n = math.ceil(l / bufSize)
   local i = 1
   local k = 1
   local subMsg = ""
   while k<=n  do
      subMsg = msg:sub(i, i+bufSize )
      sock:sendto( tostring(k)..":"..tostring(n)..":"..subMsg, addr, port)
      k = k + 1
      i = i + bufSize
   end
end

tunSend =
   function (tun, stats, buf, bufLen)
      stats.prx = stats.prx + 1
      stats.rx = stats.rx + bufLen
      local ret,err = tun:send(buf)
      if not ret then
         io.write("err:"..err.."\n")
      end
      return ret
   end

checkBuf =
   function(tun)
      local tunCfg = tunnels[tun]
      local seq = tunCfg.sequence
      local current = socket.gettime()
      --io.write("CB"..seq.bufLen..":"..seq.bufMax.."\n")
      if seq and seq.bufMax>0 then
         local stats = tunCfg.stats
         local buf = seq.bufs.next
         while buf and ((buf.rxSeq == (seq.rxSeq+1) % 2^32) or (seq.bufLen > seq.bufMax) or (seq.bufTimer < current - buf.time)) do
            --io.write("BL:"..tostring(seq.bufLen > seq.bufMax).." TO:"..tostring(seq.bufTimer < current -buf.time).."\n")
           if seq.bufTimer < current -buf.time then
              --io.write("TO:"..seq.bufTimer..":"..current..":"..buf.time.."\n")
           end
            tunSend(buf.tun, tunCfg.stats, buf.buf, buf.bufLen)
            seq.rxSeq = buf.rxSeq
            seq.bufLen = seq.bufLen - 1
            seq.bufs.next = seq.bufs.next.next
            buf = seq.bufs.next
         end
         if seq.bufLen>0 then
            tunnels.Bs[tun] = tunCfg
         else
            tunnels.Bs[tun] = nil
         end
      end
   end

insertBuf =
   function(seq, tun, stats, buf, bufLen, rxSeq)
      --io.write(rxSeq.."Buf\n")
      local c = seq.bufs
      while c.next and (c.next.rxSeq-seq.rxSeq+2^32)%2^32 < (rxSeq-seq.rxSeq+2^32)%2^32 do
         c = c.next
      end
      if (c.next) and (c.next.rxSeq == rxSeq) then
         -- repeated seq, so discard
         stats.eprx = stats.eprx + 1
      else
         local new = {
            next = c.next,
            tun = tun,
            buf = buf,
            bufLen = bufLen,
            rxSeq = rxSeq,
            time = socket.gettime()
         }
         c.next = new
         seq.bufLen = seq.bufLen + 1
         checkBuf(tun)
      end
   end

syncBuf =
   function(seq, tun, stats, buf, bufLen, rxSeq)
      --io.write("SyncBuf\n")
      local c
      local current = socket.gettime()
      if seq.syncLen > 0 then
         c = seq.syncs
         while c.next do
            if (current - c.next.time > seq.syncTimer) then
               c.next = c.next.next
               stats.eprx = stats.eprx + 1
               seq.syncLen = seq.syncLen - 1
            else
               c = c.next
            end
         end
      end
      if seq.syncLen >= seq.syncMax then
         c = seq.syncs
         c.next = c.next.next
         stats.eprx = stats.eprx + 1
         seq.syncLen = seq.syncLen - 1
      end

      if buf then
         c = seq.syncs
         while c.next and c.next.rxSeq < rxSeq do
            c = c.next
         end
         if (c.next) and (c.next.rxSeq == rxSeq) then
            -- repeated seq, use last
            c.next.buf = buf
            c.next.bufLen = bufLen
            stats.eprx = stats.eprx + 1
         else
            local new = {
               next = c.next,
               tun = tun,
               buf = buf,
               bufLen = bufLen,
               rxSeq = rxSeq,
               time = socket.gettime()
            }
            c.next = new
            seq.syncLen = seq.syncLen + 1
         end
      end
      --io.write(seq.syncLen..":"..seq.syncMax.."\n")
      if seq.syncLen >= seq.syncMax then
         c = seq.syncs
         local i = 0
         local from = c.next.rxSeq
         local to = from + seq.syncMax
         local p = c.next
         while p do
            if p.rxSeq>=from and p.rxSeq<=to then
               i = i + 1
            end
            p = p.next
         end
        
         if i >= seq.syncIntegrity*seq.syncMax/100 then
            io.write("Resync Sequence Number\n")
            c = seq.syncs
            seq.rxSeq = ((c.next.rxSeq -1) + 2^32) % 2^32
            local rxSeq, tun, buf, bufLen, stats
            while c.next do
               rxSeq = c.next.rxSeq
               tun = c.next.tun
               buf = c.next.buf
               bufLen = c.next.bufLen
               stats = tunnels[tun].stats

               if rxSeq == ((seq.rxSeq + 1) % (2^32)) then
                  tunSend(tun, stats, buf, bufLen)
                  seq.rxSeq = rxSeq
               else
                  --io.write("Cur:"..rxSeq.." P:"..seq.rxSeq.."\n")
                  if (( rxSeq - seq.rxSeq) + (2^32)) % (2^32) > (2^31)
                  then --out of order silently discard it
                     stats.eprx = stats.eprx + 1
                  else
                     if seq.bufMax>0 then
                        insertBuf(seq, tun, stats, buf, bufLen, rxSeq)
                     else
                        tunSend(tun, stats, buf, bufLen)
                        seq.rxSeq = rxSeq
                     end
                  end
               end
               seq.syncLen = seq.syncLen - 1
               c.next = c.next.next
            end
         end
      end
   end

function main(conf, port)
   local p = io.popen("cat /proc/sys/net/ipv4/ip_default_ttl")
   defaultTtl = ((p:read("*a") or 64):match("([%d]+)") or 64)
   p:close()
   local udp = socket.udp()
   udp:sendto("free()", "127.0.0.1", port) -- kill prev instance
   udp:close()
   socket.sleep(1)
   ser = require"serialize"
   local gre, tun, err
   createCTL("127.0.0.1", port)
   io.write("conf="..ser.serialize(conf, 1).."\n")
   for k,v in pairs(conf or {}) do
      if (tostring(k)):match("^gre[%d]+$") then
         io.write("Creating tunnel:"..k.."; ")
         err = controlCommands.addTunnel(k, v)
         io.write("result:"..err.."\n")
      end
   end
   local time
   local TA = 1
   while true do
      local rcv,send,err
      --io.write(tostring(time).."\n")
      --      io.write("TA:"..TA.."\n")
      rcv, send, err = socket.select(tunnels, {} , TA)
      current = socket.gettime()
      if current > tunnels.KATime then
         local next = 10
         for tun,tunCfg in pairs(tunnels.KAs or {}) do
            if current - tunCfg.KALastTime > tunCfg.KAPeriod then
               sendKeepAlive(tun)
               tunCfg.KALastTime = current
            end
            diff = tunCfg.KAPeriod - (current - tunCfg.KALastTime)
            diff = diff >1/1000 and diff or 1/1000
            next = (next > diff) and diff or next
         end
         TA = TA > next and next or TA
         tunnels.KATime = current + next
      end
      --io.write("CT:"..current.." T:"..tunnels.BTime.."\n")

      --this routine is twice repeated
      if current > tunnels.BTime then
         local next = 60
         for tun,tunCfg in pairs(tunnels.Bs or {}) do
            local buf = tunCfg.sequence.bufs.next
            if buf then
               if current - buf.time > tunCfg.sequence.bufTimer then
                  --io.write("BT:"..buf.time.."\n")
                  checkBuf(tun)
               end
               buf = tunCfg.sequence.bufs.next
               if buf then
                  diff = tunCfg.sequence.bufTimer - (current - buf.time)
                  diff = diff >1/1000 and diff or 1/1000
                  next = (next > diff) and diff or next
               end
            end
         end
         TA = TA > next and next or TA
         tunnels.BTime = current + next
      end
     
      local sockCfg
      local sock
      local readTime = 1 / (#rcv + 1) -- we want read all select-ed threads for no more then 1 second
      for k,v in ipairs(rcv) do
         local threadReadStart = socket.gettime()
         local threadReadCount = 0
         sock = v
         sockCfg = tunnels[v]
         --       io.write(sockCfg.name)
         if sockCfg.type == "tun" then
            local stats = sockCfg.stats
            local buf, ip
            local tun = sock
            local tunCfg = sockCfg
            local gre = tunCfg.link
            local greCfg = tunnels[gre]
            ip = greCfg.source
            repeat
               buf = tun:receive()
               threadReadCount = threadReadCount + 1
               if (buf and #buf or 0) >= 20 then
                  if (tunCfg.encap == 2048) and (tunCfg.tos == 'inherit') then
                     gre:setoption("ip-tos", buf:byte(2) or 0)
                  else
                     gre:setoption("ip-tos", tonumber(tunCfg.tos) or 0)
                  end
                  gre:setoption("ip-ttl", tonumber(tunCfg.ttl) or defaultTtl)
                  buf = encapGRE(buf, tunCfg, nil, tunCfg.sequence and tunCfg.sequence.txSeq)
                  local ret, err = gre:sendto(buf, tunCfg.destination, 0)
                  if tunCfg.sequence then
                     tunCfg.sequence.txSeq = tunCfg.sequence.txSeq + 1
                  end
                  if not ret then
                      stats.eptx = stats.eptx + 1
                  else
                     stats.tx = stats.tx + #buf + 20
                     stats.ptx = stats.ptx + 1
                  end
               end
            until ((buf and #buf or 0) <= 0) or ((socket.gettime()-readTime) > threadReadStart)
         elseif sockCfg.type == "GRE" then
            local buf, encap, ip, key, checksum
            local gre = sock
            local greCfg = sockCfg
            local tun
            local tunCfg
            repeat
               buf, ip = v:receivefrom()
               if (buf and #buf or 0) > 0 then
                  local bufLen = #buf
                  buf, encap, checksum, key, rxSeq = decapGRE(buf)
                  if buf then
                     tun = tunnels.hash[greCfg.device]
                     tun = tun and tun[greCfg.source]
                     tun = tun and tun[ip] and tun[ip][key]
                     if tun then
                        tunCfg = tunnels[tun]
                        local stats = tunCfg.stats
                        if (encap == 0) and (rxSeq == nil) then
                           --my keepalive
                           stats.lkaprx = stats.lkaprx + 1
                           stats.lkarx = stats.lkarx + bufLen
                           --keepalive
                           tunCfg.KAWait = false
                           tunCfg.KAError = 0
                           --io.write("keepalive received\n")
                           if  not tunCfg.KAUP then
                              tunCfg.KAUP = true
                              io.write(tunCfg.name.." up\n")
                              os.execute("ifconfig "..tunCfg.name.." up")
                              --io.write("KAE:"..tunCfg.KAError.."KAW:"..tostring(tunCfg.KAWait).."\n")
                           end
                        elseif encap == 2048 and buf:byte(10) == 47 and buf:byte(23) == 0 and buf:byte(24) == 0  and (rxSeq == nil)
                        then
                           -- keepalive from neigborg
                           local tos = buf:byte(2) or 0
                           local ttl = buf:byte(9) or defaultTtl
                           local IPHeaderLen = 4 * (buf:byte(1) - 4 * 16)
                           buf = buf:sub(1 + IPHeaderLen)
                           stats.rkaprx = stats.rkaprx + 1
                           stats.rkarx = stats.rkarx + bufLen
                           --io.write("sending keepAlive Reply\n")
                           
                           gre:setoption("ip-tos", tos)
                           gre:setoption("ip-ttl", ttl)
                           local ret,err = gre:sendto(buf, tunCfg.destination, 0)
                           if not ret then
                              stats.erkaptx = stats.erkaptx + 1
                           else
                              stats.rkatx = stats.rkatx + 20 + #buf
                              stats.rkaptx = stats.rkaptx + 1
                           end
                           
                           --tunnels.KAGRE:send(buf)
                        elseif  encap == tunCfg.encap then
                           if tunCfg.sequence and rxSeq then
                              local seq = tunCfg.sequence
                              
                              if rxSeq == ((seq.rxSeq + 1) % (2^32)) then
                                 tunSend(tun, stats, buf, bufLen)
                                 seq.rxSeq = rxSeq
                              else
                                 --io.write("Cur:"..rxSeq.." P:"..seq.rxSeq.."\n")
                                 if (( rxSeq - seq.rxSeq) + (2^32)) % (2^32) > (2^31)
                                 then --out of order silently discard it
                                    if seq.syncMax>0 then
                                       syncBuf(seq, tun, stats, buf, bufLen, rxSeq)
                                    else
                                       stats.eprx = stats.eprx + 1
                                    end
                                 else
                                    if seq.bufMax>0 then
                                       insertBuf(seq, tun, stats, buf, bufLen, rxSeq)
                                    else
                                       tunSend(tun, stats, buf, bufLen)
                                       seq.rxSeq = rxSeq
                                    end
                                 end
                              end
                           else
                              tunSend(tun, stats, buf, bufLen)
                           end
                        else
                           -- io.write("Bad enc:"..encap.."\n")
                           stats.eprx = stats.eprx + 1
                        end
                     else
                        tunnels.stats.enfprx = tunnels.stats.enfprx + 1
                     end
                  else
                     tunnels.stats.ebpprx = tunnels.stats.ebpprx + 1
                  end
               end
            until ((buf and #buf or 0) <= 0) or ((socket.gettime()-readTime) > threadReadStart)
            threadReadStart = socket.gettime()
            threadReadCount = 0
         elseif sockCfg.type == "CTL"  then
            local CTL = sock
            local CTLCfg = sockCfg
            local f, res, err
            local mes, addr, port = CTL:receivefrom()
            if mes then
               if addr == "127.0.0.1" and (
                  mes:find("^free%(%)$") or
                  mes:find("^config%(%)$") or
                  mes:find("^delTunnel%(['\"]gre%d+['\"]%)$") or
                  mes:find("^statistic%(['\"]gre%d+['\"]%)$") or
                  mes:find("^statistic%(%)$") or
               mes:find("^addTunnel%(['\"]gre%d+['\"][,]{[%[%]-a-zA-Z=0-9\"'_,.{}]+}%)$")
         )
               then
                  io.write("IN CMD:"..mes.."\n")
                  f, err = loadstring("return controlCommands."..mes)
                  if f then
                     res, err = f()
                  else
                     err = "Error in command"
                  end
               else
                  err = "Access denied"
               end
               io.write("OUT CMD:"..tostring( res or err ).."\n")
               splitSend(CTL, tostring( res or err ), addr, port)
               --CTL:sendto( tostring( res or err ) , addr, port)
            end
         end
      end
      
      current = socket.gettime()
      --this routine is twice repeated
      if current > tunnels.BTime then
         local next = 10
         for tun,tunCfg in pairs(tunnels.Bs or {}) do
            local buf = tunCfg.sequence.bufs.next
            if buf then
               if current - buf.time > tunCfg.sequence.bufTimer then
                  --io.write("BT:"..buf.time.."\n")
                  checkBuf(tun)
               end
               buf = tunCfg.sequence.bufs.next
               if buf then
                  diff = tunCfg.sequence.bufTimer - (current - buf.time)
                  diff = diff >0 and diff or 0
                  next = (next > diff) and diff or next
               end
            end
         end
         TA = TA > next and next or TA
         tunnels.BTime = current + next
      end
   end
end
io.write("Starting GRE daemon\n")
main(loadfile(arg[1])(), tonumber(arg[2]) or 50009)

