module(..., package.seeall)
local s = require'scheme'
local cosmo = require'cosmo'

local pamConfPath = "/etc/pam.d/"
local pamLibPath = "/usr/lib/security/"
local service2name = {
  login = "login",
  ssh = "sshd",
}

local radiusServerConfFilePrefix = "/etc/raddb/server_"
local function pamLib(tag, conf)
  local parm = {
    deny = function()
      return "pam_deny.so"
    end,
   ["local"] = function()
      return "pam_unix.so shadow nullok"
    end,
    radius = function()
      local fname = radiusServerConfFilePrefix..tag
      local f = io.open(fname, 'w')
      if f then
        f:write(
          ("%s:%d %s %d"):format(
            conf.server, 
            conf.port, 
            tostring(conf.secret), 
            conf.timeout
          )
        )
        
        f:close()
      end
      return "pam_radius_auth.so conf="..fname
    end,
    tacacs = function()
      return ("pam_tacplus.so encrypt server=%s secret=%s timeout=%d"):format(
        conf.server, tostring(conf.secret), conf.timeout
      )
    end,
  }
  return parm[conf.type]()
end

local template = "$do_cards[[$facil sufficient /usr/lib/security/$lib $parms\n]]"
cards = {
  {facil = "auth"},
  {facil = "account"}, 
  {facil = "password"}, 
  {facil = "session"}, 
}

local function createPamFile(service, conf)
  local fname = pamConfPath..(service2name[service] or service)
  local f = io.open(fname, 'w')
  if not f then return false end
  local text = {}
  local ufpstr
  local ufp = function(facil)
    if facil == 'auth' then
      if ufpstr then return ufpstr end
      ufpstr = "use_first_pass"
    end
    return''
  end
  for i = 0, #conf do
    if conf[i] then
      text[#text+1] = cosmo.f(template){
        do_cards = function()
          for _, v in ipairs(cards) do
            cosmo.yield{
              facil = v.facil, 
              lib = pamLib(service..i, conf[i]), 
              parms = ufp(v.facil)
            }
          end
        end
      }
    end
  end
  f:write(table.concat(text))
  f:close()
  return true
end

local aaaAply = function(path, sch, conf, prevConf)
  local conf = s.addDefault(path, sch, conf, true)
  os.execute("rm "..radiusServerConfFilePrefix..'*')
  for k, _ in sch:_pairs() do
    if k~='default' then
      if not createPamFile(
        k, conf[k] and conf[k][1] and conf[k] or conf.default
      ) then return false end
    end
  end
  return true
end

local aaaconf = s.newScheme{
  _H = "aaaParms",
  ["type"] = s.newScheme{
    _H = "aaaParmsType",
    _T = 'choice',
    _A = {"local", "radius", "tacacs", "deny"},
    _V = s.choiceVal,
    _D = "local",
    _P = 100,
    _DEP = 1,
  },

  server = function(_, path)
    local typepar = C.getConfOrDefault(s.clonePath(path, -1, "type"))
    return (typepar == 'radius' or typepar == 'tacacs') and 
    s.ipAddressType{
      _H = "aaaParmsServer",
      _D = "127.0.0.1",
    }
  end,
  port = function(_, path)
    local typepar = C.getConfOrDefault(s.clonePath(path, -1, "type"))
    return (typepar == 'radius') and 
    s.tcpPortType{
      _H = "aaaParmsPort",
      _D = 1812,
    }
  end,
  secret = function(_, path)
    local typepar = C.getConfOrDefault(s.clonePath(path, -1, "type"))
    return (typepar == 'radius' or typepar == 'tacacs') and 
    s.stringType{
      _H = "aaaParmsSecret",
    }
  end,
  timeout = function(_, path)
    local typepar = C.getConfOrDefault(s.clonePath(path, -1, "type"))
    return (typepar == 'radius' or typepar == 'tacacs') and 
    s.integerType{
      _H = "aaaParmsTimeout",
      _A = {1, 60*60},
      _D = 3,
    }
  end,
}

aaaScheme = s.newScheme{
  _H = "aaa",
  _Z = aaaAply,

  default = s.newScheme{
  _H = "aaaDefault",
    [0] = aaaconf,
    _LIST = aaaconf
  },

  login = s.newScheme{
    _H = "aaaLogin",
    _LIST = aaaconf
  },

  ppp = s.newScheme{
    _H = "aaaPpp",
    _LIST = aaaconf
  },

  ssh = s.newScheme{
    _H = "aaaSsh",
    _LIST = aaaconf
  },
}
