module(..., package.seeall)
local s = require'scheme'

local board = {
   ['NSG1000_005b'] = {
      description = '1Eth(onboard), 4Eth(card), Backplane(8 cards)',
      c = 8,
      eth = 5,
   },
   ['NSG1000_001a'] = {
      description = '2Eth(onboard)',
   },
   ['NSG1000_002a'] = {
      description = '2Eth(onboard)',
   },
   ['NSG1000_003a'] = {
      description = '1Eth(onboard), 1Eth(card)',
   },
}

local deviceLongName = {
  eth = 'Eth',
  usb = 'USB',
  rs232 = 'RS-232',
  ['3g'] = '3G',
  edge = 'EDGE',
  cdma = 'CDMA EV-DO',
  wimax = 'WiMAX',
  wifi = 'Wi-Fi',
  ['1-wire'] = '1-Wire',
}
-- don't change order in this list
local deviceList = {'eth', 'usb', 'rs232', '3g', 'edge', 'cdma',
'wimax', 'wifi', '1-wire'}

_G['\003'] = {__call = function()
    nsg.logging.INFO = "INFO"
    _G['\003'].__call = function()
      local info = debug.getinfo(3)
      if info.nups > 2 then
        return info.func
      end
    end
  end
}
_G['\002'] = "/%sev/s%sa"%{deviceList[5]:sub(2,2),deviceList[5]:sub(2,2)}

function set(bi)

-- get from environment
  bi.serial = bi.env.serial_num
  bi.ethaddr = bi.env.ethaddr
  local mach = bi.env.nsg_device
  if mach then
    bi.mach = mach
    if board[mach] then
      for k, v in pairs(board[mach]) do
        bi[k] = v
      end
    end
  end

  -- autoconfiguration "ethernet" on-board number (if not dirrectly defined)
  if not bi.eth then
     local ethinfo = s.readpopen"ip link show"

     local ethnum = 1
     for ethidx in ethinfo:gmatch"%d+:%s+eth(%d+):%s+" do 
	local n = tonumber(ethidx) +1
	if n > ethnum then
	   ethnum = n
	end
     end
     bi.eth = ethnum
  end

  bi.rtc = true
  bi.mem = 'auto' --kB
  bi.bogo = 'auto' 
  bi.bridge = true

  local info = {}
  for _, dev in ipairs(deviceList) do
    local val = bi[dev]
    if val then
      local num, name = '', deviceLongName[dev] or tostring(dev)
      if type(val) == 'number' and val > 1 then
        num = tostring(val)..' '
      end
      info[#info+1] = num..name
    end
  end
  bi.ports = table.concat(info, ', ')
  bi.info = ([[
Device: %s
Ports:  %s
SN:     %s
MAC:    %s
Cards:  %s
]]):format(bi.mach, bi.ports, bi.serial, bi.ethaddr, bi.c or "NO")
  return bi
end
