local io = io
local pairs = pairs
local ipairs = ipairs
local type = type
local tostring = tostring
local table = table
module(...)

local board = {
   ['NSG700_4AU_v5'] = {
      description = '1Eth(3), s1(async/sdsl/usb), s2(usb), 3Async(RS-232)',
      eth = 1,
      s1 = true,
      s2 = true,
      a = 3, 
   },
   ['NSG700_4AU_v6'] = {
      description = '1Eth(3), s1(async/sync/usb), s2(async/usb), 3Async(RS-232), 1-Wire',
      eth = 1,
      s1 = true,
      s2 = true,
      a = 3,
      ["1-wire"] = true,
   }, 
   ['NSG700_4AU_v6m'] = {
      description = '1Eth(3), s1(async/sync/sdsl/usb), s2(async/usb), 3Async(RS-232), 1-Wire',
      eth = 1,
      s1 = true,
      s2 = true,
      a = 3,
      ["1-wire"] = true,
   },
   ['NSG700_4AU_v6f'] = {
      description = '1Eth(2+optic), s1(async/sync/sdsl/usb), 2Async(RS-232)',
      eth = 1,
      s1 = true,
      a = 2,
   },
   ['NSG700_4AU_v7'] = {
      description = '1Eth(3), 1Eth(1), s1(async/sync/sdsl/usb), s2(async/usb), 3Async(RS-232), 1-Wire',
      eth = 2,
      s1 = true,
      s2 = true,
      a = 3,
      ["1-wire"] = true,
   },
   ['NSG700_8AU'] = {
      description = '1Eth(3), s1(async/sdsl/usb), 7Async(RS-232)',
      eth = 1,
      s1 = true,
      a = 7, 
   },
   ['NSG700_16AU'] = {
      description = '1Eth(3), s1(async/sdsl/usb), 11Async(RS-232)',
      eth = 1,
      s1 = true,
      a = 11, 
   },
}

local deviceLongName = {
   eth = 'Ethernet(10/100)',
   a = 'Async RS-232',
   s1 = 'Interface module (slot1)',
   s2 = 'Interface module (slot2)',
   ['1-wire'] = '1-Wire',
}
local deviceList = {'eth', 'a', 's1', 's2', '1-wire', }

function set(bi)
--[[
   bi.info = "NSG-700  under construction :)"
   bi.eth = 1
   bi.s1 = true
   bi.s2 = true
   bi["1-wire"] = true
   bi.a = 3  -- a1, a2, a3
   return bi  
]]--
  bi.serial = bi.env.serial_num
  bi.ethaddr = bi.env.ethaddr
  local mach = bi.env.nsg_device
  if mach then
    bi.mach = mach
    if board[mach] then
      for k, v in pairs(board[mach]) do
        bi[k] = v
      end
    end
  end
  bi.rtc = true
  bi.mem = 'auto' --kB
  bi.bogo = 'auto'
  bi.bridge = true
  if bi.eth > 1 then
    local t = {{}}
    for i = 0, bi.eth - 1 do
      t[1]["eth"..i] = true
    end
    bi.invalidBridgeGroup = t
  end

  local info = {}
  for _, dev in ipairs(deviceList) do
    local val = bi[dev]
    if val then
      local num, name = '', deviceLongName[dev] or tostring(dev)
      if type(val) == 'number' and val > 1 then
        num = tostring(val)..' '
      end
      info[#info+1] = num..name
    end
  end
  bi.ports = table.concat(info, ', ')
  bi.info = ([[
Device: %s
Ports:  %s
SN:     %s
MAC:    %s
]]):format(bi.mach, bi.ports, bi.serial, bi.ethaddr)
  return bi

end
