function RootOrAdminScheme(sch)
  return function(_, path)
    return (C.session and 
      (C.session.user == 'root' or C.session.rights == 'admin')
    ) and 
    sch
  end
end

local s = require'scheme'
--local ppp = require'nsgpppconf'
local sys = require'nsgsysconf'
local ser = require'nsgserconf'
local tools = require'nsgtoolsconf'
local oneWire = require'nsg1wireconf'
local rs232 = require'nsgrs232conf'
local ipnod = require'nsgipconf'
local eth = require'nsgethconf'
local port = require'nsgportconf'
local WiFi = require'nsgwifi'
local ipsec = require'nsgipsecconf'
local uitcp = require'nsguitcpconf'
--local nf = require'nsgnfconf'
local pppoe = require'nsgpppoeconf'
local pptp = require'nsgpptpconf'
local stunnel = require'nsgstunnel'
local gre = require'nsggre'
local usb = require'nsgusbconf'
local search = require'nsgsearch'

local portListScheme = s.newScheme{
  _H = "portList",
  ["3g"] = BI['3g'] and function(_, path) 
    C.setOption(path, 'wirelessType', "3g")
    C.setOption(path, 'simNumber', "2")
    C.setOption(path, 'controlledPower', "pin")
    C.setOption(path, 'smsType', "aux")
    return port.portScheme
  end,
  edge = BI.edge and function(_, path) 
    C.setOption(path, 'wirelessType', "edge")
    C.setOption(path, 'simNumber', "2")
    C.setOption(path, 'controlledPower', "pin")
    C.setOption(path, 'smsType', "+++")
    return port.portScheme
  end,

  cdma = BI.cdma and function(_, path) 
    C.setOption(path, 'wirelessType', "cdma")
    C.setOption(path, 'simNumber', "1")
    C.setOption(path, 'controlledPower', "pin")
    return port.portScheme
  end,

  ["1-wire"] = BI["1-wire"] and function(_, path) 
    usb.probeUSBDeviceScheme(nil, path, 2) -- second iface !
    oneWire.setDevName()
    return oneWire.oneWireScheme
  end,

  ["rs-232"] = BI.rs232 and rs232.rs232Scheme,
  eth0 = eth.ethScheme{_H = "portEth0"},
  eth1 = BI.eth and BI.eth > 1 and eth.ethScheme{_H = "portEthN"},
  eth2 = BI.eth and BI.eth > 2 and eth.ethScheme{_H = "portEthN"},
  eth3 = BI.eth and BI.eth > 3 and eth.ethScheme{_H = "portEthN"},
  eth4 = BI.eth and BI.eth > 4 and eth.ethScheme{_H = "portEthN"},
  wlan0 = BI.wifi and WiFi.WiFiScheme,

  s1 = BI.s1 and  usb.portTemplateFunction,   

  s2 = BI.s2 and  usb.portTemplateFunction,   

  m1 = BI.m1 and  usb.portTemplateFunction,   

  a1 = BI.a and BI.a > 0 and function(_, path)
    usb.port2tty["a1"] = "/dev/ttyS2" -- TEMPO!
    return rs232.rs232Scheme
  end,
  a2 = BI.a and BI.a > 1 and function(_, path)
    usb.port2tty["a2"] = "/dev/ttyS3" -- TEMPO!
    return  rs232.rs232Scheme
  end,
  a3 = BI.a and BI.a > 2 and function(_, path)
    usb.port2tty["a3"] = "/dev/ttyS4" -- TEMPO!
    return  rs232.rs232Scheme
  end,
  usb = BI.usb and usb.portTemplateFunction,


  ["c1-1"] =  BI.c and BI.c > 0 and usb.portTemplateFunction,
  ["c1-2"] =  BI.c and BI.c > 0 and usb.portTemplateFunction,
  ["c2-1"] =  BI.c and BI.c > 1 and usb.portTemplateFunction,
  ["c2-2"] =  BI.c and BI.c > 1 and usb.portTemplateFunction,
  ["c3-1"] =  BI.c and BI.c > 2 and usb.portTemplateFunction,
  ["c3-2"] =  BI.c and BI.c > 2 and usb.portTemplateFunction,
  ["c4-1"] =  BI.c and BI.c > 3 and usb.portTemplateFunction,
  ["c4-2"] =  BI.c and BI.c > 3 and usb.portTemplateFunction,
  ["c5-1"] =  BI.c and BI.c > 4 and usb.portTemplateFunction,
  ["c5-2"] =  BI.c and BI.c > 4 and usb.portTemplateFunction,
  ["c6-1"] =  BI.c and BI.c > 5 and usb.portTemplateFunction,
  ["c6-2"] =  BI.c and BI.c > 5 and usb.portTemplateFunction,
  ["c7-1"] =  BI.c and BI.c > 6 and usb.portTemplateFunction,
  ["c7-2"] =  BI.c and BI.c > 6 and usb.portTemplateFunction,
  ["c8-1"] =  BI.c and BI.c > 7 and usb.portTemplateFunction,
  ["c8-2"] =  BI.c and BI.c > 7 and usb.portTemplateFunction,

}

local tunnelListScheme = s.newScheme{
  _H = "tunnelList",
  ipsec = ipsec.ipsecScheme,
  pppoe = pppoe.pppoeScheme,
  pptp = pptp.pptpScheme,
  uitcp = uitcp.uitcpScheme,
  stunnel = stunnel.stunnelScheme,
  gre = gre.greScheme,
}

local ethernetScheme = s.newScheme{
   _H = "ethernet",
   bridge = BI.bridge and eth.bridgeScheme,
   switch = (
      BI.arch == 'nsg600' and BI.eth and BI.eth > 1 and
         eth.ethernetSwitchScheme or
         BI.arch == 'nsg700' and eth.ethernetSwitchRtlScheme
   ),
   bond = eth.bondScheme,
}

local function fixConfigPath(from, to)
   local oldConf = C.getConfByPath(from)
   if oldConf and table.next(oldConfig) then
      C.log_warn("Absolute configuration detected at:{"..table.concat(from,".").."}")
      C.setConfigValue(to, oldConf)
      C.setConfigValue(from, nil)
      local info = "Configuration moved from:{"..table.concat(from,".").."} to:{"..table.concat(to,".").."},  check them and save!"
      C.log_warn(info)
      return false, info
   end
   return true
end

local fixConfigs = {
   [{"bridge"}] = {"ethernet","bridge"},
   [{"ethernet-switch"}] = {"ethernet","switch"},
}

return s.newScheme{
  _H = "root",
  _Z = function()
          local ret = true
          local err = {}
          local res,msg
          for from,to in pairs(fixConfigs) do
             res, msg = fixConfigPath(from, to)
             ret = ret and res
             err[#err + 1] = msg
          end
          return ret, not ret and table.concat(err,"\n") or nil
       end,
  ip = ipnod.ipScheme,
  port = portListScheme,
  system = sys.systemScheme,
  services = ser.servicesScheme,
  tools = tools.toolsScheme,
  tunnel = tunnelListScheme,
  search = search.searchHelpScheme,
  ethernet = ethernetScheme,
  ["user-defined"] = function(_, path)
    local f, e = loadfile'/etc/userdefinedscheme'
    if not f then
      if not e:match"^cannot open" then
        C.log_error(e)
      end
    else
      local res, sch = pcall(f)
      if res then
        return sch
      else
        C.log_error(sch)
      end
    end
  end,
  --[[ --Example of /etc/userdefinedscheme file

local s = require'scheme'

local help = require'nsgconfhelp'
-- для имен хелпов использовать префикс userdefined, 
-- чтобы избежать переопределения
help.userdefined = "User defined scheme"
help.userdefinedLsetc = "Show /etc directory"

return s.newScheme{
  _H = "userdefined",
  ['ls-etc'] = s.newScheme{
    _H = "userdefinedLsetc",
    _X = function()
      return('\n'..s.readpopen("ls -l /etc/".." 2>&1"))
    end,
    _R = 'user',
  },
}

]]
}
