module(..., package.seeall)
local s = require'scheme'
local ppp = require'nsgpppconf'
--------------------------
-- common prio (high)
local PPTPCOMMONADMSTATEPRIO = 100 
local PPTPCOMMONMODEPRIO     = 90

-- client prio
local PPTPCLIENTDESTINATIONPRIO = 50 
local PPTPCLIENTSOURCEIPPRIO    = 40
local PPTPCLIENTKEEPALIVEPRIO   = 30 
local PPTPCLIENTRETRYPRIO       = 20 


-- server prio
local PPTPSERVERIPADDRESSPOOLPRIO   = 50
local PPTPSERVERLIMITPRIO           = 40
local PPTPSERVERLISTENPRIO          = 30

--common prio (low)
local PPTPCOMMONPPPPRIO      = 10
local PPTPCOMMONSHOWPRIO     = 5
--------------------------
local pptpOneShow = s.newScheme {
   _H = "pptpOneShow",

   ["log"] =  s.newScheme{
      _H = "pptpOneShowLog",
      _X = function(path)
	      path = s.clonePath(path, -2)
	      return C.getDaemonLog(path)
	   end,
      _R = 'user',
   },  

}
--------------------------
local function AddOptions(target, opt, value) 
   if value then
      target[#target + 1] = opt
      target[#target + 1] = value
   end
end

local function restartPptp(path, sch, conf, prevConf)
   local conf = s.addDefault(path, sch, conf, true) --clone conf with defaults
   local ApCmdStart = {}
   local TmpFileList = {}
   --clone PPP-conf with defaults (only MAIN)
   local pppconf = s.addDefault(path, sch, conf, true) 
   local mainCom = ppp.getPPPParams(path, pppconf, 'main', {})
   local ConfText = table.concat(mainCom, '\n')..'\n'
   local filename = os.tmpname()
   local f = io.open(filename, 'w')
   if f then
      f:write(ConfText)
      f:close()
   end
   TmpFileList[#TmpFileList+1] = filename   --- ppp-parameters file
   
   if conf.mode and conf.mode == "server" then
      local pptpfilename = os.tmpname()
      local f = io.open(pptpfilename, 'w')
      if f then
	 if conf["ip-address-pool"].from then
	    f:write("localip " .. conf["ip-address-pool"].from ..'\n')
	 end
	 if conf["ip-address-pool"].to then
	    f:write("remoteip " .. conf["ip-address-pool"].to ..'-254\n')
	 end
	 f:close()
      end      
      TmpFileList[#TmpFileList+1] = pptpfilename   --- pptpd-parameters file
      
      ApCmdStart = {'/sbin/pptpd', '-f',  }
      AddOptions(ApCmdStart, '-l', conf.listen)
      AddOptions(ApCmdStart, '-C', conf.limit)
      AddOptions(ApCmdStart, '-c', pptpfilename)
      AddOptions(ApCmdStart, '-o', filename)
   else
      if not conf.destination then
	 return true, "Destination address needed to start!"
      end
      ApCmdStart = {'/usr/sbin/pptp', '--debug', '--sync', conf.destination, 'nodetach', 'sync', 'file', filename }
      AddOptions(ApCmdStart, '--localbind', conf["source-ip"])
      AddOptions(ApCmdStart, '--idle-wait', conf.keepalive)
      AddOptions(ApCmdStart, '--max-echo-wait', conf.keepalive and conf.retry and (conf.keepalive * conf.retry))
   end
   
   
   if conf["adm-state"] == 'up' then
      C.restartDaemon(path, ApCmdStart, { tmpFileList = TmpFileList, atExit = C.increasingSleep})
      return true
   end
   C.restartDaemon(path)
   return true
end

--------------------------
local pptpOne = s.newScheme {
   
   _Z = restartPptp,
   
   _H = "pptpOne",
   
   _N = function(sch, conf)
	   local need = {
	      true,
	      mode = true,
	      name = true,
	      ppp = true,
	      show = true,
	      ["adm-state"] = true,
	   }
	   if conf and conf.mode == "server" then
	      need["limit"] = true
	      need["listen"] = true
	      need["ip-address-pool"] = true
	   else
	      need["destination"] = true
	      need["source-ip"] = true
	      need["keepalive"] = true
	      need["retry"] = true
	   end
	   return need
	end,
   
   -- Common Lines 
   
   mode = s.newScheme {
      _H = "pppoxOneMode",
      _T = 'choice',
      _A = {"client", "server"},
      _V = s.choiceVal,
      _D = "client",
      _DEP = 1,
      _P = PPTPCOMMONMODEPRIO,
   },
   
   ["adm-state"] = s.newScheme{
      _T = "choice",
      _A = {"up", "down"},
      _V = s.choiceVal,
      _H = "pppoxOneAdmState",
      _D = "down",
      _P = PPTPCOMMONADMSTATEPRIO,
   },

   ppp = ppp.pppScheme{_P = PPTPCOMMONPPPPRIO, },

   show = pptpOneShow{_P = PPTPCOMMONSHOWPRIO, },

-- Pptp-server params
   
   listen = s.ipAddressType { _H = "pptpOneListen", _D = "127.0.0.1", _P = PPTPSERVERLISTENPRIO, },
   
   limit = s.integerType { _H = "pptpOneLimit", _A = {1, 100}, _D = 10, _P = PPTPSERVERLIMITPRIO, },

   ["ip-address-pool"] = s.IpAddressPool{ _P = PPTPSERVERIPADDRESSPOOLPRIO, },   

-- Pptp-client params

   destination = s.ipAddressType { _H = "pptpOneDestination", _P = PPTPCLIENTDESTINATIONPRIO, },

   ["source-ip"] = s.ipAddressType { _H = "pptpOneSourceIp", _P = PPTPCLIENTSOURCEIPPRIO, },

   keepalive =  s.integerType{
      _H = "pptpOneKeepalive",
      _A = {1, 3600},
      _P = PPTPCLIENTKEEPALIVEPRIO,
      _DEP = 1,
   },

   retry = function(_, path)
	      local keepalive = C.getConfOrDefault(s.clonePath(path, -1, "keepalive"))
	      return keepalive and s.integerType{ 
		 _H = "pptpOneRetry", 
		 _A = {1, 100}, 
		 _P = PPTPCLIENTRETRYPRIO,
	      }
	   end,   
}

pptpScheme = s.newScheme {
   _H = "pptpScheme",

   _NAME = pptpOne,
}
