module(..., package.seeall)
local s = require'scheme'
local eth = require'nsgethconf'
local port = require'nsgportconf'
local iface = require'nsgifaceconf'
local rs232 = require'nsgrs232conf'
local nsg = require'nsg'
local WiFi = require'nsgwifi'
local sens = require'nsgsensconf'

local lfs = require "lfs"
local socket = require "socket"

local BASE_DIR = "/sys/bus/usb/devices"

local allTypes = {"eth", "cdma", "edge", "3g", "wifi", "2com", "rs-232", "multisensor", "empty", }

local modules = { 
  ["0b05/1706"] = { descr = "Wifi-Dongle", type = "wifi", }, 
  ["0bda/8189"] = { descr = "WiFi-Module RTL", type = "wifi", },
  ["18e8/6229"] = { descr = "WiFi-Module RaLink", type = "wifi", },
  ["07a6/1986"] = { descr = "um-et100",    type = "eth", },
  ["0b95/772a"] = { descr = "um-et100-v2", type = "eth", },
  ["16d8/5543"] = { descr = "uim-evdo",    type = "cdma", },
  ["16d8/680a"] = { descr = "uim-evdo",    type = "cdma", },
  ["0471/1234"] = { descr = "uim-edge-v3", type = "edge", },
  ["0471/0020"] = { descr = "uim-edge-v3", type = "edge", },
  ["05c6/6000"] = { descr = "uim-3g",      type = "3g", },
  ["05c6/9000"] = { descr = "um-3g",       type = "3g", },
  ["0403/6010"] = { descr = "um-2COM",     type = "2com", childs = {".0", ".1"}},
  ["0483/fffc"] = { descr = "ms-6",        type = "multisensor", },
}
-------------------------------------------------
local function writeToFile(fname, str)
  local f, err = io.open(fname, 'w')
  if f then
    f:write(str) 
    f:close()
  else
    --log_error("Can't open "..fname.." "..err)
  end

end
------------------------------------
local NSG1800_Position2Port = {   
  ["1"] = {
    port = { "s2", },
    powerOn = function ()
      writeToFile('/sys/class/gpio/gpio137/value', '0')
    end,
    powerOff = function ()
      writeToFile('/sys/class/gpio/gpio137/value', '1')
    end,
    setGpio = function ()
      writeToFile('/sys/class/gpio/export', '137')   -- DTR s2
      writeToFile('/sys/class/gpio/gpio137/direction', 'out')
      writeToFile('/sys/class/gpio/export', '140')   -- "RTS" s2
      writeToFile('/sys/class/gpio/gpio140/direction', 'out')
    end,
    setSim = function (sim)
      writeToFile('/sys/class/gpio/gpio140/value', sim) 
    end,
    blacklist = {"rs-232", },
  },
  ["2"] = {
    port = { "s1", },
    powerOn = function ()
      writeToFile('/sys/class/gpio/gpio217/value', '0')
    end,
    powerOff = function ()
      writeToFile('/sys/class/gpio/gpio217/value', '1')
    end,
    setGpio = function ()
      writeToFile('/sys/class/gpio/export', '217')   -- DTR s1
      writeToFile('/sys/class/gpio/gpio217/direction', 'out')
      writeToFile('/sys/class/gpio/export', '223')   -- "RTS" s1
      writeToFile('/sys/class/gpio/gpio223/direction', 'out')
    end,
    setSim = function (sim)
      writeToFile('/sys/class/gpio/gpio223/value', sim) 
    end,
    blacklist = {"rs-232", },
  },
  ["3"] = {
    port = { "m1", },
    powerOn = function () 
      writeToFile('/sys/class/gpio/gpio134/value', '0')
    end,
    powerOff = function () 
      writeToFile('/sys/class/gpio/gpio134/value', '1')
    end,
    setGpio = function ()
      writeToFile('/sys/class/gpio/export', '134')   -- power control pin (PA6)
      writeToFile('/sys/class/gpio/gpio134/direction', 'out')
    end,
    setSim = function (sim)
      -- now unimplemented yet
    end,
    blacklist = {"rs-232", "multisensor", "2com", "eth",},
  },
}
local NSG600_Position2Port = {   
  ["1"] = {
    port = {"usb", },
    powerOn = function () 
      writeToFile('/sys/class/gpio/gpio15/value', '0')
    end,
    powerOff = function () 
      writeToFile('/sys/class/gpio/gpio15/value', '1')
    end,
    setGpio = function ()
      writeToFile('/sys/class/gpio/export', '15')
      writeToFile('/sys/class/gpio/gpio15/direction', 'out')
    end,
  },
  ["2"] = {
    port = { "3g", "edge", "cdma", },
    powerOn = function () 
      writeToFile('/sys/class/gpio/gpio9/value', '1')
    end,
    powerOff = function () 
      writeToFile('/sys/class/gpio/gpio9/value', '0')
    end,
    setGpio = function ()
      writeToFile('/sys/class/gpio/export', '9')
      writeToFile('/sys/class/gpio/export', '12')
      writeToFile('/sys/class/gpio/gpio9/direction', 'out')
      writeToFile('/sys/class/gpio/gpio12/direction', 'out')
    end,
    setSim = function (sim)
      writeToFile('/sys/class/gpio/gpio12/value', sim)
    end,
  },
}
local NSG700_Position2Port = {   
  ["1.1"] = {
    port = {"s1",  },
    powerOn = function ()
      writeToFile('/sys/class/gpio/gpio66/value', '0')
    end,
    powerOff = function ()
      writeToFile('/sys/class/gpio/gpio66/value', '1')
    end,
    setGpio = function ()
      writeToFile('/sys/class/gpio/export', '66')   -- PIO_PB2 (64 + 2)
      writeToFile('/sys/class/gpio/gpio66/direction', 'out')
      writeToFile('/sys/class/gpio/export', '53')   -- PIO_PA21 (32 + 21)
      writeToFile('/sys/class/gpio/gpio53/direction', 'out')
    end,
    setSim = function (sim)
      writeToFile('/sys/class/gpio/gpio53/value', sim)   -- ???
    end,
  },
  ["1.2"] = {
    port = {"usb", },
    powerOn = function () 
    end,
    powerOff = function () 
    end,
    setGpio = function ()
    end,
  },
  ["1.3"] = {
    port = {"s2", },
    powerOn = function () 
      nsg.modem(port2tty["s2-aux"], "DTR", 1)
    end,
    powerOff = function () 
      nsg.modem(port2tty["s2-aux"], "DTR", 0)
    end,
    setGpio = function ()
      local path = {"s2-aux", } -- path emulation
      probeUSBDeviceScheme(nil, path, 1) -- first iface ! (set real tty to table)
    end,
    setSim = function (sim)
		-- sim = '1' - main (then RTS=Off (0)) 
		-- sim = '0' - aux  (then RTS=On  (1))	
		nsg.modem(port2tty["s2-aux"], "RTS", (1 - tonumber(sim)))
    end,
  },
  ["1.4"] = {
    port = {[1] = "s2-aux", [2] = "1-wire", },
    powerOn = function () 
    end,
    powerOff = function () 
    end,
    setGpio = function ()
    end,
    setSim = function (sim)
    end,
  },
}
local NSG1000_Position2Port = {   
  ["1.1"] = {
    port = {"c1-2", },
    powerOn = function () end,
    powerOff = function () end,
    setGpio = function ()  end,
    blacklist = {"cdma", "edge", "3g", "wifi", "2com", "rs-232", "multisensor", },
  },
  ["1.2"] = {
    port = {"c2-2", },
    powerOn = function () end,
    powerOff = function () end,
    setGpio = function ()  end,
    blacklist = {"cdma", "edge", "3g", "wifi", "2com", "rs-232", "multisensor", },
  },
  ["1.3"] = {
    port = {"c3-2", },
    powerOn = function () end,
    powerOff = function () end,
    setGpio = function ()  end,
    blacklist = {"cdma", "edge", "3g", "wifi", "2com", "rs-232", "multisensor", },
  },
  ["1.4"] = {
    port = {"c4-2", },
    powerOn = function () end,
    powerOff = function () end,
    setGpio = function ()  end,
    blacklist = {"cdma", "edge", "3g", "wifi", "2com", "rs-232", "multisensor", },
  },
  ["2.1"] = {
    port = {"c1-1", },
    powerOn = function () end,
    powerOff = function () end,
    setGpio = function ()  end,
    blacklist = {"cdma", "edge", "3g", "wifi", "2com", "rs-232", "multisensor", },
  },
  ["2.2"] = {
    port = {"c2-1", },
    powerOn = function () end,
    powerOff = function () end,
    setGpio = function ()  end,
    blacklist = {"cdma", "edge", "3g", "wifi", "2com", "rs-232", "multisensor", },
  },
  ["2.3"] = {
    port = {"c3-1", },
    powerOn = function () end,
    powerOff = function () end,
    setGpio = function ()  end,
    blacklist = {"cdma", "edge", "3g", "wifi", "2com", "rs-232", "multisensor", },
  },
  ["2.4"] = {
    port = {"c4-1", },
    powerOn = function () end,
    powerOff = function () end,
    setGpio = function ()  end,
    blacklist = {"cdma", "edge", "3g", "wifi", "2com", "rs-232", "multisensor", },
  },
  ["3.1"] = {
    port = {"c5-2", },
    powerOn = function () end,
    powerOff = function () end,
    setGpio = function ()  end,
    blacklist = {"cdma", "edge", "3g", "wifi", "2com", "rs-232", "multisensor", },
  },
  ["3.2"] = {
    port = {"c6-2", },
    powerOn = function () end,
    powerOff = function () end,
    setGpio = function ()  end,
    blacklist = {"cdma", "edge", "3g", "wifi", "2com", "rs-232", "multisensor", },
  },
  ["3.3"] = {
    port = {"c7-2", },
    powerOn = function () end,
    powerOff = function () end,
    setGpio = function ()  end,
    blacklist = {"cdma", "edge", "3g", "wifi", "2com", "rs-232", "multisensor", },
  },
  ["3.4"] = {
    port = {"c8-2", },
    powerOn = function () end,
    powerOff = function () end,
    setGpio = function ()  end,
    blacklist = {"cdma", "edge", "3g", "wifi", "2com", "rs-232", "multisensor", },
  },
  ["4.1"] = {
    port = {"c5-1", },
    powerOn = function () end,
    powerOff = function () end,
    setGpio = function ()  end,
    blacklist = {"cdma", "edge", "3g", "wifi", "2com", "rs-232", "multisensor", },
  },
  ["4.2"] = {
    port = {"c6-1", },
    powerOn = function () end,
    powerOff = function () end,
    setGpio = function ()  end,
    blacklist = {"cdma", "edge", "3g", "wifi", "2com", "rs-232", "multisensor", },
  },
  ["4.3"] = {
    port = {"c7-1", },
    powerOn = function () end,
    powerOff = function () end,
    setGpio = function ()  end,
    blacklist = {"cdma", "edge", "3g", "wifi", "2com", "rs-232", "multisensor", },
  },
  ["4.4"] = {
    port = {"c8-1", },
    powerOn = function () end,
    powerOff = function () end,
    setGpio = function ()  end,
    blacklist = {"cdma", "edge", "3g", "wifi", "2com", "rs-232", "multisensor", },
  },
}
-------------------------------------------------

local  Position2Port = 
((BI.arch == 'nsg600') and NSG600_Position2Port) or
((BI.arch == 'nsg700') and NSG700_Position2Port) or 
((BI.arch == 'nsg1000') and NSG1000_Position2Port) or 
((BI.arch == 'nsg1800') and NSG1800_Position2Port) or {}


local function FindPositionByPort(port)
  for pos, v in pairs(Position2Port) do           --- for all BUS-position 
    for _, name in ipairs(v.port) do             --- for all aliases of the current port
      if name == port then
        return pos
      end
    end
  end
  return nil
end 
-------------------------------------------------
isWirelessPort = {
  ["3g"] = true,
  ["edge"] = true,
  ["cdma"] = true,
}

port2tty = {
  ["3g"] = "/dev/ttyUSB0",
  ["edge"] = "/dev/ttyACM0",
  ["cdma"] = "/dev/ttyACM0",
  ["rs-232"] = "/dev/ttyAM0",
  ["1-wire"] =  "/dev/ttyAM1",   --++ static for NSG-600 (not autodetected, overwrite in NSG-700, )
}
setmetatable(port2tty, {__index = function()return''end})

port2netdev = {
  ["wlan0"] = "wlan0",  -- compatibility to NSG-600 (port "wlan0")
}
-------------------------------------

local function IsDevice(name) 
  if name == "." or 
  name == ".." or 
  string.find(name, "usb") or
  string.find(name, ":") 
  then
    return false
  end
  return true
end

local function GetValue (dir, name)
  local f = io.open(dir .. name)
  if f then
    return tostring(f:read())
  end
  return "NONE"
end
-------------------------------------
local function SetFeature(module, name)
  local head = name .. ":" 
  module.netdevice = nil
  module.ttydevice = module.childs and {}  
  -- .ttydevice = nil (if no childs)
  -- .ttydevice = {}  (if there are childs)
  for entry in lfs.dir(BASE_DIR) do
    if (string.find(entry, head)) then
      local iface = BASE_DIR .. "/" .. entry
      for fname in lfs.dir(iface) do
        if (string.find(fname, "tty")) then
          local ttyname = string.gsub(fname, "^tty:", "")
          if module.childs then
            module.ttydevice[#module.ttydevice + 1] = ttyname -- set sequential
          else
            module.ttydevice = module.ttydevice or ttyname -- set only first !!!
          end
        end
        if (string.find(fname, "net")) then
          module.netdevice = string.gsub(fname, "^net:", "")
        end
      end
    end
  end
end
-------------------------------------
function FindDevice(port)
  for de in lfs.dir(BASE_DIR) do
    local name = de
    if IsDevice(name) then
      local devpath = string.gsub(name, "^%d%-", "")
      local pos = FindPositionByPort(port)
      if devpath == pos then
        local devdir = BASE_DIR .. "/" .. name .. "/"
        local vidpid =  GetValue(devdir, "idVendor") .. "/" .. GetValue(devdir, "idProduct")
        if modules[vidpid] then
          SetFeature(modules[vidpid], name)
          return modules[vidpid]
        else
          return {} -- to return "unknown" device {} when we haven't this vid/pid
        end
      end
    end
  end
  -- to return "nil"  when NO device in the slot 
end
------------------------------------
local um2comScheme = function(_, path)
  local name = path[#path]
  return s.newScheme {
    [name..'.0'] = rs232.rs232Scheme,
    [name..'.1'] = rs232.rs232Scheme,
  }			
end

------------------------------------
local type2scheme = {
  ["eth"] = { 
     scheme = eth.ethScheme{
	_Z = function(path, sch, conf) 
		probeUSBDeviceScheme(nil, path)
		return eth.ethScheme._Z(path, sch, conf) -- iface.interfaceMainApply(path, sch, conf) 
	     end, 
     },
  },
  ["edge"] = {
     scheme = port.portScheme,
     options = { ['wirelessType'] = "edge", ['simNumber'] = "2", ["smsType"]= "+++",},
  },
  ["cdma"] = { 
     scheme = port.portScheme,
     options = { ['wirelessType'] = "cdma", ['simNumber']= "1", },
  },
  ["3g"] = {
     scheme = port.portScheme,
     options = { ['wirelessType'] = "3g", ['simNumber'] = "2", ["smsType"]= "aux",},
  }, 
  ["2com"] = {
     scheme = um2comScheme,
  },

  ["wifi"] = {
     scheme = WiFi.WiFiScheme,
  },
  
  ["rs-232"] = {
     scheme = rs232.rs232Scheme{
	_Z = function(path, sch, conf) 
		port2tty[path[#path]] = "/dev/ttyS1" -- TEMPO
		C.setOption(path, 'controlledPower', nil)
		C.setOption(path, 'simNumber', "2")
		return rs232.rs232Scheme._Z(path, sch, conf) 
	     end,
     },
  },

  ["multisensor"] = {
     scheme = sens.ms6Scheme,
  },

  ["empty"] = {
     scheme = s.newScheme {},
  }
}

------------------------------------
function probeUSBDeviceScheme(_, path, interface)
  local portName = path[#path]
  local device = FindDevice(portName)

  if device then
    if(device.netdevice) then
      iface.addIfaceAlias( portName, device.netdevice)
      port2netdev[portName] = device.netdevice
    end
    if(device.ttydevice) then
      if device.childs then   -- this device has multiple interfaces
        if interface then    -- if iface is set 
          -- get this 'ttydevice' (if present) and NO use a "suffix" 
          port2tty[portName] = device.ttydevice[interface] and ('/dev/' .. device.ttydevice[interface])
        else
          for i, v in ipairs(device.childs) do -- through all "childs" and use "suffix"
            if device.ttydevice[i] then
              port2tty[portName .. device.childs[i]] = '/dev/' .. device.ttydevice[i]
            end
          end
        end
      else  -- NO "childs" NO "suffix"
        port2tty[portName] = '/dev/' .. device.ttydevice
        isWirelessPort[portName] = true
      end
    else
      port2tty[portName] = nil
      isWirelessPort[portName] = nil
    end
    return device.type
  end
end
------------------------------------
local function SetPrivateOptions(path, type)
   for opt, val in pairs(type2scheme[type].options or {}) do
      C.setOption(path, opt, val)
   end
end
------------------------------------
local delayedApply
delayedApply = function(path, sch, conf, delay)
			local type = probeUSBDeviceScheme(nil, path)
			if type then
			   SetPrivateOptions(path, type)
			   return C.applyConf(path, nil, true) -- Apply with recursive!
			else
			   local nextdelay = delay + 3
			   if delay < 30 then	
			      C.log_info("*** Device (%s) not ready, trying after %d sec ", path[#path], nextdelay)		      
			      C.timer:add(nextdelay, delayedApply, path, sch, conf, nextdelay) -- Apply after
			      return true
			   else
			      C.log_info("DelayApply Not done *** name = %s (no device)", path[#path])
			   end
			   return true
			end
		     end
------------------------------------
local function allowModules(path) 
   local pos = FindPositionByPort(path[#path])
   local allowTypes = {} 
   if pos then
      for _, type in pairs(allTypes) do
	 local allow = true
	 for _, v in pairs(Position2Port[pos].blacklist or {}) do
	    if v == type then
	       allow = nil
	       break
	    end
	 end
	 allowTypes[#allowTypes + 1] = allow and type 
      end
   else
      allowTypes[#allowTypes + 1] = "empty"
   end
   return allowTypes
end
------------------------------------
function portTemplateFunction(_, path)
  local type = C.getConfByPath(s.clonePath(path, nil, "type")) or 'empty'

  C.setOption(path, 'controlledPower', "pin")

  local portTemplate = {
     _H = "portTemplate",

     _Z = function(path, sch, conf) 
	     if conf and conf.type ~= "empty" then 
		local realtype = probeUSBDeviceScheme(nil, path)
		if realtype == conf.type then
		   SetPrivateOptions(path, realtype)
		   return  type2scheme[type].scheme._Z(path, sch, conf)  --- Apply real
		else
		   local pos = FindPositionByPort(path[#path])  
		   if pos then 
		      Position2Port[pos].setGpio()
		      Position2Port[pos].powerOn()  --- PowerON
		      C.timer:add(3, delayedApply, path, sch, conf, 3) -- Apply after
		   end
		end
	     end
	     return true
	  end,

     _P = -10, -- autodetected (removable) ports are below "fixed" in menu
     type = function(_, path)
	       return s.choiceType{
		  _H = "portType",
		  _A = allowModules(s.clonePath(path, -1)),
		  _D = "empty",
		  _P = 99,
		  _V = function (...)
			  local type = s.choiceVal(...)
			  if type and type2scheme[type] then
			     local path = s.clonePath(path, -1)
			     SetPrivateOptions(path, type)
			  end
			  return  s.choiceVal(...)
		       end,
		  _DEP = 1, 
	       }
	    end,

    update = s.newScheme{
      _H = "portUpdate",
      _X = function(path, sch, val)
        local path = s.clonePath(path, -1)
        local pos = FindPositionByPort(path[#path])
        if pos then 
          Position2Port[pos].setGpio()
          -- C.threadSleep(1)
          Position2Port[pos].powerOn()
        end
        local newtype = probeUSBDeviceScheme(nil, path) or 'empty'
        C.setConfigValue(path, { type = newtype, } )
	SetPrivateOptions(path, newtype)
        C.cleanDaemonLog(path)
        return
      end,
      _P = 99,
      _DEP = 1,
    },

  }

  return type2scheme[type or 'empty'].scheme(nil, path)(portTemplate)
end
------------------------------------
function getOptionRecursive(path, option)
  local opt = C.getOption(path, option)
  local path = path
  if opt then
    return opt
  end
  path = s.clonePath(path, -1)
  if #path == 0 then
    return nil
  else
    return getOptionRecursive(path, option)
  end
end
------------------------------------

setGPIO = 
function(path) 
  local pos = FindPositionByPort(path[#path])
  return pos and Position2Port[pos].setGpio()
end
------------------------------------

togglePower = 
function(path)
  local pos = FindPositionByPort(path[#path])
  if pos then
    Position2Port[pos].powerOff()
    C.threadSleep(1)
    Position2Port[pos].powerOn()
  end
end 
------------------------------------
mainSimNumber = '1'
auxSimNumber = '0'

setSim = 
function (path, n)
  local pos = FindPositionByPort(path[#path])
  C.log_info("Set SIM %s(%d)", n == mainSimNumber and "MAIN" or "AUX", n)
  return pos and Position2Port[pos].setSim(n)
end
------------------------------------
powerOff = 
function(path)
  local pos = FindPositionByPort(path[#path])
  if pos then
    Position2Port[pos].powerOff()
  end
end
